//! This class parses one line of the vCard/iCalendar format at a time.

constant NAME = 1;
constant PARAMS = 2;
constant VALUE = 3;
.ContentLine current_line = .ContentLine();
int crlf = 0;
int pos = 0;
string buf = "";
int state = NAME;
string current_key;

void unfold(int start) {

    while (-1 != (start = search(buf, "\r\n", start))) {
	if (start+2 >= sizeof(buf)) return;

	switch (buf[start+2]) {
	case 9:
	case 32:
	    // unfold
	    if (sizeof(buf) == start+3) {
		buf = buf[0..start-1];
	    } else {
		buf = buf[0..start-1] + buf[start+3..];
	    }
	}
	start += 2;
    }
}

void flush() {
    if (sizeof(buf) && pos > 0) {
	if (pos == sizeof(buf)) {
	    buf = "";
	    pos = 0;
	} else if (pos < sizeof(buf)) {
	    buf = buf[pos..];
	    pos = 0;
	} else error("bad position in buf.\n");
    }
}

// used for param-value
string|array(string) param_value() {
    array(string) a = ({});
    int tpos = pos;

LOOP: while (1) {
	if (buf[tpos] == '"') {
	    int i = search(buf, '"', tpos+1); // we are allowing CTLS here... maybe bad

	    if (-1 == i || i >= sizeof(buf)-1) { // on the end we have to try again
		return 0;
	    }

	    if (buf[i+1] == ';' || buf[i+1] == ':') {
		a += ({ buf[tpos+1..i-1] });
		pos = i+2;
		break;
	    } else if (buf[i+1] == ',') {
		a += ({ buf[tpos+1..i-1] });
		tpos = i+2;
	    } else error("misplaced quoted string.\n");
	} else {
	    int i;
	    for (i = tpos;i < sizeof(buf); i++) switch (buf[i]) {
	    case ',':
		a += ({ buf[tpos..i-1] });
		tpos = i+1;
		continue LOOP;
	    case ':':
	    case ';':
		a += ({ buf[tpos..i-1] });
		pos = i+1;
		break LOOP;
	    case 0x00..0x08:
	    case 0x0A..0x1F:
	    case 0x7F:
	    case 0x22: // "
		string t = ""+buf[i];
		error("forbidden char %s in name part.\n", t);
	    }
	    return 0;
	}
    }

    return sizeof(a) > 1 ? a : a[0];
}

string parse_name() {
    int c, i;
    for (i = pos;i < sizeof(buf); i++) switch (c = buf[i]) {
    case ':': // end of name, beginning of value
    case ';': // end of name, beginning of params
    case '=':
	string ret = buf[pos..i-1];
	pos = i+1;
	return ret;
    case 'a'..'z':
    case 'A'..'Z':
    case '0'..'9':
    case '-':
    case '.': // group character
	break;
    default:
	error("forbidden char in name part at %O.\n", buf[i..]);
    }

    return 0;
}

//! Add data to the internal buffer. Will be processed on parse().
void add_data(string data) {
    if (sizeof(data)) {
	flush();

	int t = sizeof(buf);

	buf += data;

	unfold(max(t-2,0));
    }
}

//! Parse at most one content-line. Returns 0 if no complete content-line
//! can be found in the buffer.
.ContentLine parse(void|string data) {
    if (data) add_data(data);

    if (state == NAME) {
	string name = parse_name();

	if (!name) return 0;
	current_line->name = name;
	if (buf[pos-1] == ':') {
	    state += 2;
	} else if (buf[pos-1] == ';') {
	    state++;
	} else error("forbidden char in name part.\n");
    }

    if (state == PARAMS) {
OUTER:	while (pos < sizeof(buf)) {
	    if (current_key) {
		array(string)|string val = param_value();

		if (!val) return 0;

		current_line->params[current_key] = val;
		current_key = 0;

		if (buf[pos-1] == ':') { 
		    state = VALUE;
		    break;
		}
	    } else {
		string key = parse_name();

		if (!key) return 0;

		current_key = key;

		switch (buf[pos-1]) {
		case ';':
		    current_line->params[current_key] = 1; 
		    current_key = 0;
		case '=': // value
		    break;
		case ':':
		    current_line->params[current_key] = 1; 
		    state = VALUE;
		    current_key = 0;
		    break OUTER;
		default:
		     error("forbidden char %s in param-name.\n", buf[pos-2..]);
		}
	    }
	}
    }

    if (state == VALUE) {
	int i;
	for (i = pos;i < sizeof(buf); i++) switch (buf[i]) {
	case '\r':
	    // Id say this format sux.. you cannot determine whether 
	    // you have incomplete data or not unless you know if more 
	    // is coming
	    // therefore we hack here.. at the end it should be safe to end a line even if
	    // not unfolded yet.
	    if (i+1 >= sizeof(buf) || i+2 >= sizeof(buf) && lower_case(current_line->name) != "end") return 0; // we dont know yet.  

	    if (buf[i+1] == '\n') {
		string val = buf[pos..i-1];
		pos = i+2;
		current_line->value = val;

		.ContentLine ret = current_line;
		current_line = .ContentLine();
		state = NAME;
		return ret;
	    }
	case 0x00..0x08:
	case 0x0A..0x0C:
	case 0x0E..0x1F:
	case 0x7F:
	    error("forbidden char in name part.\n");
	}
    }

}
