constant BEGIN = 0;
constant PROP  = 1;
constant COMP_BEGIN  = 2;
constant COMP  = 3;
constant END = 4;

int state = BEGIN;
object current_object;
object current_parser;
.LineParser line_parser = .LineParser();
program|object|mapping(string:object) blueprint;
object|mapping(string:object) parsers;

object get_new_object(string name) {
    object o;

    if (mappingp(blueprint) || objectp(blueprint)) {
	program p = blueprint[name];

	if (p) o = p();
    } else if (programp(blueprint)) {
	o = blueprint();
    }

    if (!o) {
	error("Could not create object for component %s.\n", name);	
    }

    return o;
}

//! Create a new Parser object.
//! @param blueprint
//! 	Use this if you want the parser to create something else than
//! 	@[Public.Standards.vCard.Object] instances.
//! @param parsers
//! 	Parsers to use for specific Components. This is used by the iCalendar
//! 	module. There are no components in vCards.
void create(program|object|mapping(string:object) blueprint, void|object|mapping(string:object) parsers) {
    this_program::parsers = parsers;
    this_program::blueprint = blueprint;
}

//! Parse at most one Object. Call this every time when more data is 
//! available. Will return a parsed object if
//! there is a complete one available. You probably want to
//! put this into a while-loop in case data contained more than
//! one object.
int|object parse(void|string data) {
    if (data) line_parser->add_data(data);

    while (.ContentLine line = line_parser->parse()) {
	.Object o = parse_cline(line);

	if (o) return o;
    }

    return 0;
}

//! Same as @[parse()] but is working on a
//! @[Public.Parser.vCard.ContentLine] instead.
int|object parse_cline(.ContentLine line) {
    string name = lower_case(line->name);
    string value = lower_case(line->value);

    switch (state) {
    case BEGIN:
	if (name == "begin") {
	    current_object = get_new_object(value);
	    current_object->name = value;
	}
	state = PROP;
	break;
    case PROP:
	if (name == "end") {
	    state = COMP_BEGIN;
	} else if (name == "begin") {
	    if (current_object->add_component && parsers) { // has components
		state++;
	    } else {
		error("Components are not allowed in %O.\n", current_object);	
	    }
	} else {
	    current_object[name] = line;
	    break;
	}
    case COMP_BEGIN:
	if (name == "begin") {
	    current_parser = parsers[value];

	    if (!current_parser) error("cannot create component for %O.\n", line->value);
	    state++;
	} else if (name == "end" && value == current_object->name) {
	    object ret = current_object;
	    current_object = .Object();
	    state = BEGIN;
	    if (ret->check_consistency) ret->check_consistency();
	    return ret;
	} else {
	    error("strange line: %O.\n", line);
	}
    case COMP:
	mixed ret = current_parser->parse_cline(line);

	if (objectp(ret)) {

	    current_object->add_component(ret);
	    current_parser = 0;
	    state = COMP_BEGIN;
	    break;
	}
    }
}

